/*
 * IRC - Internet Relay Chat, ircd/m_check.c
 * Copyright (C) 1990 Jarkko Oikarinen and
 * University of Oulu, Computing Center
 *
 * See file AUTHORS in IRC package for additional names of
 * the programmers.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: m_check.c,v 1.6 2003/03/01 17:57:20 cometen Exp $
 */


#include "channel.h"
#include "check.h"
#include "client.h"
#include "hash.h"
#include "ircd.h"
#include "ircd_alloc.h"
#include "ircd_defs.h"
#include "ircd_features.h"
#include "ircd_reply.h"
#include "ircd_string.h"
#include "ircd_snprintf.h"
#include "list.h"
#include "listener.h"
#include "match.h"
#include "numeric.h"
#include "numnicks.h"
#include "querycmds.h"
#include "send.h"
#include "s_user.h"
#include "s_debug.h"

#include <string.h>


int m_check(struct Client *cptr, struct Client *sptr, int parc, char *parv[])
{
   struct Channel *chptr;
   struct Client *acptr;
   char *param;
   int showchan = 0;

   if (parc < 2)
   {
      send_reply(sptr, ERR_NEEDMOREPARAMS, "CHECK");
      return 0;
   }

   /* This checks to see if -c has been supplied for hostmask checking */
   if (parc >= 3 && (!strcmp(parv[2],"-c")))
   {
      showchan = 1;
   }

   if (IsChannelName(parv[1]))  /* channel */
   {
      if ((chptr = FindChannel(parv[1])))
      {
         checkChannel(sptr, chptr);
         checkUsers(sptr, chptr);
      }
      else
      {
         send_reply(sptr, ERR_SEARCHNOMATCH, "CHECK");
      }
   }
   else if ((acptr = FindClient(parv[1])) && !(FindServer(parv[1])))  /* client and not a server */
   {
      if (!IsRegistered(acptr))
   	{
         send_reply(sptr, ERR_SEARCHNOMATCH,"CHECK");
         return 0;
      }

    checkClient(sptr, acptr);
   }
   else if ((acptr = FindServer(parv[1])))  /* server */
   {
      checkServer(sptr, acptr);
   }
   else if (checkHostmask(sptr, parv[1], showchan) > 0)  /* hostmask */
   {
      return 1;
   }
   else  /* no match */
   {
      send_reply(sptr, ERR_SEARCHNOMATCH, "CHECK");
   }
 
  return 1;

}

static int checkClones(struct Channel *chptr, char *nick, char *host)
{
   int clones = 0;
   struct Membership *lp;
   struct Client *acptr;

   for (lp = chptr->members; lp; lp = lp->next_member) 
   {
      acptr = lp->user;
      if (!strcmp(acptr->cli_user->host, host) && strcmp(acptr->cli_name, nick))
      {
         /* this is a clone */
         clones++;
      }
   }

   /* this is not a clone */
   return ((clones) ? clones + 1 : 0);
}

void checkUsers(struct Client *sptr, struct Channel *chptr)
{
   struct Membership *lp;
   struct SLink *slp;
   struct Client *acptr;

   char outbuf[BUFSIZE], ustat[64];
   int cntr = 0, opcntr = 0, vcntr = 0, clones = 0, bans = 0, c = 0;

   send_reply(sptr, RPL_DATASTR, "Users (@ = op, + = voice, * = clone)");

   for (lp = chptr->members; lp; lp = lp->next_member)
   {

      acptr = lp->user;

      if ((c = checkClones(chptr, acptr->cli_name, acptr->cli_user->host)) != 0)
      {
         ircd_snprintf(0, ustat, sizeof(ustat), "%2d ", c);
         clones++;
      }
      else
      {
         strcpy(ustat, "   ");
      }

      if (chptr && is_chan_op(acptr, chptr))
      {
         strcat(ustat, "@");
         opcntr++;
      }
      else if (chptr && has_voice(acptr, chptr))
      {
         strcat(ustat, "+");
         vcntr++;
      }
      else
      {
         strcat(ustat, " ");
      }

      send_reply(sptr, RPL_CHANUSER, ustat, acptr->cli_name, acptr->cli_user->username,
          acptr->cli_user->host, acptr->cli_info, (IsAccount(acptr) ? acptr->cli_user->account : ""));

      cntr++;
   }

   send_reply(sptr, RPL_DATASTR, " ");

   ircd_snprintf(0, outbuf, sizeof(outbuf),
      "Total users:: %d (%d ops, %d voiced, %d clones)",
      cntr, opcntr, vcntr, clones);
   send_reply(sptr, RPL_DATASTR, outbuf);

   send_reply(sptr, RPL_DATASTR, " ");

   /* Bans */
   send_reply(sptr, RPL_DATASTR, "Bans on channel::");

   for (slp = chptr->banlist; slp; slp = slp->next)
   {
      ircd_snprintf(0, outbuf, sizeof(outbuf),  "[%d] - %s - Set by %s, on %s", 
         ++bans, slp->value.ban.banstr, slp->value.ban.who, ctime(&slp->value.ban.when));
      send_reply(sptr, RPL_DATASTR, outbuf);
   }

   if (bans == 0)
   {
      send_reply(sptr, RPL_DATASTR, "<none>");
   }

   send_reply(sptr, RPL_ENDOFCHECK, " ");
}

void checkChannel(struct Client *sptr, struct Channel *chptr)
{
   char outbuf[TOPICLEN + MODEBUFLEN + 64], modebuf[MODEBUFLEN], *s;
   unsigned int is_limit = 0, is_keyed = 0;

   /* Header */
   send_reply(sptr, RPL_DATASTR, " ");
   send_reply(sptr, RPL_CHKHEAD, "channel", chptr->chname);
   send_reply(sptr, RPL_DATASTR, " ");

   /* Creation Time */
   ircd_snprintf(sptr, outbuf, sizeof(outbuf), "  Creation time:: %s", ctime(&chptr->creationtime));
   send_reply(sptr, RPL_DATASTR, outbuf);

   /* Topic */
   if (strlen(chptr->topic) <= 0)
   {
      send_reply(sptr, RPL_DATASTR, "          Topic:: <none>");
   }
   else 
   {
      ircd_snprintf(sptr, outbuf, sizeof(outbuf), "          Topic:: %s", chptr->topic);
      send_reply(sptr, RPL_DATASTR, outbuf);

      /* ..set by */
      ircd_snprintf(sptr, outbuf, sizeof(outbuf), "         Set by:: %s", chptr->topic_nick);
      send_reply(sptr, RPL_DATASTR, outbuf);
   }

   /* Channel Modes */

   strcpy(outbuf, "Channel mode(s):: ");

   s = modebuf;

   *s++ = '+';

   if (chptr->mode.mode & MODE_SECRET)
   {
      *s++ = 's';
   }

   if (chptr->mode.mode & MODE_PRIVATE)
   {
      *s++ = 'p';
   }

   if (chptr->mode.mode & MODE_MODERATED)
   {
      *s++ = 'm';
   }

   if (chptr->mode.mode & MODE_NOPRIVMSGS)
   {
      *s++ = 'n';
   }

   if (chptr->mode.mode & MODE_TOPICLIMIT)
   {
      *s++ = 't';
   }

   if (chptr->mode.mode & MODE_INVITEONLY)
   {
      *s++ = 'i';
   }

   if (chptr->mode.mode & MODE_NOCOLOUR)
   {
      *s++ = 'c';
   }

   if (chptr->mode.mode & MODE_NOCTCP)
   {
      *s++ = 'C';
   }

   if (chptr->mode.mode & MODE_OPERONLY)
   {
      *s++ = 'O';
   }

   if (chptr->mode.mode & MODE_ACCONLY)
   {
      *s++ = 'M';
   }


   if (chptr->mode.limit)
   {
      *s++ = 'l';
      is_limit = 1;
   }

   if (*chptr->mode.key)
   {
      *s++ = 'k';
      is_keyed = 1;
   }

   *s++ = '\0';

   if (strlen(modebuf) <= 1)
   {
      strcat(outbuf, "<none>");
   }
   else
   {
      strcat(outbuf, modebuf);
   }

   /* Key and Limit handling.. */

   if (is_limit)
   {
      ircd_snprintf(0, outbuf + strlen(outbuf), sizeof(outbuf), " %d", chptr->mode.limit);
   }

   if (is_keyed)
   {
      ircd_snprintf(0, outbuf + strlen(outbuf), sizeof(outbuf), " %s", chptr->mode.key);
   }

   send_reply(sptr, RPL_DATASTR, outbuf);

   /* Send 'END OF CHECK' message */

   send_reply(sptr, RPL_ENDOFCHECK, " ");
}

void checkClient(struct Client *sptr, struct Client *acptr)
{
   struct Channel *chptr;
   struct Membership *lp;
   char outbuf[BUFSIZE];
   time_t nowr;

   /* Header */
   send_reply(sptr, RPL_DATASTR, " ");
   send_reply(sptr, RPL_CHKHEAD, "user", acptr->cli_name);
   send_reply(sptr, RPL_DATASTR, " ");

   /* Basic WHOIS info. */
   if (MyUser(acptr))
   {
      ircd_snprintf(0, outbuf, sizeof(outbuf),  "      Signed on:: %s", ctime(&acptr->cli_firsttime));
      send_reply(sptr, RPL_DATASTR, outbuf);
   }

   ircd_snprintf(0, outbuf, sizeof(outbuf), "           Nick:: %s (%s%s)", acptr->cli_name, NumNick(acptr));
   send_reply(sptr, RPL_DATASTR, outbuf);

   ircd_snprintf(0, outbuf, sizeof(outbuf), "  User/Hostmask:: %s@%s (%s)", acptr->cli_user->username, acptr->cli_user->host,
   ircd_ntoa((const char*) &acptr->cli_ip));
   send_reply(sptr, RPL_DATASTR, outbuf);

   ircd_snprintf(0, outbuf, sizeof(outbuf), "      Real Name:: %s", cli_info(acptr)); 
   send_reply(sptr, RPL_DATASTR, outbuf); 
 

   if (IsService(acptr) == -1)
   {
      send_reply(sptr, RPL_DATASTR, "         Status:: Network Service");
   }
   else if (IsAnOper(acptr))
   {
      send_reply(sptr, RPL_DATASTR, "         Status:: IRC Operator");
   }

   ircd_snprintf(0, outbuf, sizeof(outbuf), "   Connected to:: %s", cli_name(acptr->cli_user->server));
   send_reply(sptr, RPL_DATASTR, outbuf);

   /* +s (SERV_NOTICE) is not relayed to us from remote servers,
    * so we cannot tell if a remote client has that mode set.
    * And hacking it onto the end of the output of umode_str is EVIL BAD AND WRONG
    * (and breaks if the user is +r) so we won't do that either.
    */

   if (strlen(umode_str(acptr)) < 1)
      strcpy(outbuf, "       Umode(s):: <none>");
   else
      ircd_snprintf(0, outbuf, sizeof(outbuf), "       Umode(s):: +%s", umode_str(acptr));
      send_reply(sptr, RPL_DATASTR, outbuf);

   if (acptr->cli_user->joined == 0)
      send_reply(sptr, RPL_DATASTR, "     Channel(s):: <none>");
   else if (acptr->cli_user->joined > 50) 
   {

      /* NB. As a sanity check, we DO NOT show the individual channels the
       *     client is on if it is on > 50 channels.  This is to prevent the ircd
       *     barfing ala Uworld when someone does /quote check Q :).. (I shouldn't imagine
       *     an Oper would want to see every single channel 'x' client is on anyway if
       *     they are on *that* many).
       */

      ircd_snprintf(0, outbuf, sizeof(outbuf), "     Channel(s):: - (total: %u)", acptr->cli_user->joined);
      send_reply(sptr, RPL_DATASTR, outbuf);
   }
   else
   {
      char chntext[BUFSIZE];
      int len = strlen("     Channel(s):: ");
      int mlen = strlen(me.cli_name) + len + strlen(sptr->cli_name);
      *chntext = '\0';

      strcpy(chntext, "     Channel(s):: ");
      for (lp = acptr->cli_user->channel; lp; lp = lp->next_channel)
      {
         chptr = lp->channel;
         if (len + strlen(chptr->chname) + mlen > BUFSIZE - 5) 
         {
            send_reply(sptr, RPL_DATASTR, chntext);
            *chntext = '\0';
            strcpy(chntext, "     Channel(s):: ");
            len = strlen(chntext);
         }
         if (IsDeaf(acptr))
            *(chntext + len++) = '-';
         if (is_chan_op(acptr, chptr))
            *(chntext + len++) = '@';
         else if (has_voice(acptr, chptr))
            *(chntext + len++) = '+';
         else if (IsZombie(lp))
            *(chntext + len++) = '!';
         if (len)
            *(chntext + len) = '\0';

         strcpy(chntext + len, chptr->chname);
         len += strlen(chptr->chname);
         strcat(chntext + len, " ");
         len++;
      }

      if (chntext[0] != '\0')
      send_reply(sptr, RPL_DATASTR, chntext);
   }

   /* If client processing command ISN'T target (or a registered
    * Network Service), show idle time since the last time we
    * parsed something.
    */
   if (MyUser(acptr) && !(IsService(acptr) == -1) && !(strCasediff(acptr->cli_name, sptr->cli_name) == 0)) 
   {
      nowr = CurrentTime - acptr->cli_user->last;
      ircd_snprintf(0, outbuf, sizeof(outbuf), "       Idle for:: %d days, %02ld:%02ld:%02ld",
         nowr / 86400, (nowr / 3600) % 24, (nowr / 60) % 60, nowr % 60);
      send_reply(sptr, RPL_DATASTR, outbuf);
   }

   /* Away message (if applicable) */
   if (acptr->cli_user->away) 
   {
      ircd_snprintf(0, outbuf, sizeof(outbuf), "   Away message:: %s", acptr->cli_user->away);
      send_reply(sptr, RPL_DATASTR, outbuf);
   }
   
   /* Send 'END OF CHECK' message */
   send_reply(sptr, RPL_ENDOFCHECK, " ");
}

void checkServer(struct Client *sptr, struct Client *acptr)
{
   char outbuf[BUFSIZE];

   /* Header */
   send_reply(sptr, RPL_DATASTR, " ");
   send_reply(sptr, RPL_CHKHEAD, "server", acptr->cli_name);
   send_reply(sptr, RPL_DATASTR, " ");

   ircd_snprintf(0, outbuf, sizeof(outbuf),  "   Connected at:: %s", ctime(&acptr->cli_serv->timestamp));
   send_reply(sptr, RPL_DATASTR, outbuf);

   ircd_snprintf(0, outbuf, sizeof(outbuf), "    Server name:: %s", acptr->cli_name);
   send_reply(sptr, RPL_DATASTR,  outbuf);

   ircd_snprintf(0, outbuf, sizeof(outbuf), "        Numeric:: %s --> %d", NumServ(acptr), base64toint(acptr->cli_yxx));
   send_reply(sptr, RPL_DATASTR, outbuf);

   /* Send 'END OF CHECK' message */
   send_reply(sptr, RPL_ENDOFCHECK, " ");
}

signed int checkHostmask(struct Client *sptr, char *hoststr, int showchan)
{
   struct Client *acptr;
   struct Channel *chptr;
   struct Membership *lp;
   int count = 0;
   char outbuf[BUFSIZE];
   char targhost[NICKLEN + USERLEN + HOSTLEN + 3], curhost[NICKLEN + USERLEN + HOSTLEN + 3];
   char nickm[NICKLEN + 1], userm[USERLEN + 1], hostm[HOSTLEN + 1];
   char *p = NULL;

   strcpy(nickm,"*");
   strcpy(userm,"*");
   strcpy(hostm,"*");

   if (!strchr(hoststr, '!') && !strchr(hoststr, '@'))
   {
      ircd_strncpy(hostm,hoststr,HOSTLEN);
   }
   else
   {
      if (p = strchr(hoststr, '@'))
      {
         *p++ = '\0';
         if (*p) ircd_strncpy(hostm,p, HOSTLEN);
      }

      /* Get the nick!user mask */
      if (p = strchr(hoststr, '!')) 
      {
         *p++ = '\0';
         if (*p) ircd_strncpy(userm,p,USERLEN);
         if (*hoststr) ircd_strncpy(nickm,hoststr,NICKLEN);
      }
      else if (*hoststr)
      {
         /* Durz: We should only do the following *IF* the hoststr has not already been
          * copied into hostm (ie. neither ! or @ specified).. otherwise, when we do
          * /quote check *.barrysworld.com - we end up with targhost as: *!*.barryswo@*.barrysworld.com
          */
         ircd_strncpy(userm,hoststr,USERLEN);
      }
   }

   /* Copy formatted string into "targhost" buffer */
   ircd_snprintf(0, targhost, sizeof(targhost),  "%s!%s@%s", nickm, userm, hostm);

   targhost[sizeof(targhost) - 1] = '\0';

   /* Note: we have to exclude the last client struct as it is not a real client
    * structure, and therefore any attempt to access elements in it would cause
    * a segfault.
    */

   for (acptr = GlobalClientList; acptr; acptr = cli_next(acptr)) 
   {
      /* Dont process if acptr is a unregistered client, a server or a ping */
      if (!IsRegistered(acptr) || IsServer(acptr))
         continue;

      if (IsMe(acptr))   /* Always the last acptr record */
         break;

      /* Copy host info into buffer */
      ircd_snprintf(0, curhost, sizeof(curhost), "%s!%s@%s", acptr->cli_name, acptr->cli_user->username, acptr->cli_user->host);

      if (match((const char *)targhost, (const char *)curhost) == 0) 
      {
         /* Show header if we've found at least 1 record */
         if (count == 0) 
         {
            /* Output header */ 
            send_reply(sptr, RPL_DATASTR, " ");
            send_reply(sptr, RPL_CHKHEAD, "host", targhost);

            send_reply(sptr, RPL_DATASTR, " ");
            ircd_snprintf(0, outbuf, sizeof(outbuf),  "%s   %-*s%-*s%s", "No.", (NICKLEN + 2 ), "Nick",
				(USERLEN + 2), "User", "Host");
            send_reply(sptr, RPL_DATASTR, outbuf);
         }

         ircd_snprintf(0, outbuf, sizeof(outbuf), "%-4d  %-*s%-*s%s", (count+1), (NICKLEN + 2),
            acptr->cli_name, (USERLEN + 2), acptr->cli_user->username, acptr->cli_user->host);
         send_reply(sptr, RPL_DATASTR, outbuf);

         /* Show channel output (if applicable) - the 50 channel limit sanity check
          * is specifically to prevent coredumping when someone lamely tries to /check
          * Q or some other channel service...
          */
         if (showchan == 1) 
         {
            if (acptr->cli_user->joined > 0 && acptr->cli_user->joined <= 50) 
            {
               char chntext[BUFSIZE];
               int len = strlen("      on channels: ");
               int mlen = strlen(me.cli_name) + len + strlen(sptr->cli_name);
               *chntext = '\0';

               strcpy(chntext, "      on channels: ");
               for (lp = acptr->cli_user->channel; lp; lp = lp->next_channel)
               {
                  chptr = lp->channel;
                  if (len + strlen(chptr->chname) + mlen > BUFSIZE - 5) 
                  {
                     send_reply(sptr, RPL_DATASTR, chntext);
                     *chntext = '\0';
                     strcpy(chntext, "      on channels: ");
                     len = strlen(chntext);
                  }
                  if (IsDeaf(acptr))
                     *(chntext + len++) = '-';
                  if (is_chan_op(acptr, chptr))
                     *(chntext + len++) = '@';
                  else if (has_voice(acptr, chptr))
                     *(chntext + len++) = '+';
                  else if (IsZombie(lp))
                     *(chntext + len++) = '!';
                  if (len)
                     *(chntext + len) = '\0';

                  strcpy(chntext + len, chptr->chname);
                  len += strlen(chptr->chname);
                  strcat(chntext + len, " ");
                  len++;
               }
               if (chntext[0] != '\0')
                  send_reply(sptr, RPL_DATASTR, chntext);

               send_reply(sptr, RPL_DATASTR, " ");
            }
         }
         count++;
      }
   }

   if (count > 0)
   {
      send_reply(sptr, RPL_DATASTR, " ");

      ircd_snprintf(0, outbuf, sizeof(outbuf), "Matching records found:: %d", count);
      send_reply(sptr, RPL_DATASTR, outbuf);

      send_reply(sptr, RPL_ENDOFCHECK, " ");
   }

   return count;
}

