/*
 * IRC - Internet Relay Chat, ircd/m_vhost.c
 * Copyright (C) 1990 Jarkko Oikarinen and
 *                    University of Oulu, Computing Center
 *
 * See file AUTHORS in IRC package for additional names of
 * the programmers.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: m_vhost.c,v 1.3 2003/02/24 23:15:51 cometen Exp $
 */

#include "channel.h"
#include "config.h"
#include "client.h"
#include "handlers.h"
#include "hash.h"
#include "ircd.h"
#include "ircd_defs.h"
#include "ircd_features.h"
#include "ircd_relay.h"
#include "ircd_reply.h"
#include "ircd_string.h"
#include "ircd_snprintf.h"
#include "ircd_alloc.h"
#include "list.h"
#include "msg.h"
#include "numeric.h"
#include "numnicks.h"
#include "querycmds.h"
#include "send.h"
#include "s_conf.h"
#include "s_user.h"
#include "s_debug.h"
#include "userload.h"
#include "patchlevel.h"

#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <arpa/inet.h>

/* 
 * m_vhost()
 * 
 * parv[0] = sender prefix   (nick)
 * parv[1] = v:line username (vusername)
 * parv[2] = v:line password (vpassword)
 *
 * Client Syntax: VHOST USERNAME PASSWORD
 * Conf Syntax:   V:USERNAME:PASSWORD:HOST.COM
 *
 * This will change the clients host to the host specific in the
 * ircd.conf. Clients CANNOT unset a VHOST, nor can they change
 * to another VHOST after changing to a VHOST. Clients will need
 * to quit and reconnect if they wish to revert back to their 
 * original host or changing to another VHOST. This code is based
 * on mbuna's S:Line code, ircu 2.10.11's HIDDEN_HOST code, and 
 * QuakeNets SETHOST code, as well as other new code. Opers can
 * see an original host/ip of a client in a WHOIS on a client who
 * has succesfully completed a VHOST.
 *
 * - SiRVu|caN
 */

int m_vhost(struct Client *cptr, struct Client *sptr, int parc, char *parv[])
{
  /* Variables */
  struct Membership *chan;
  struct ConfItem* aconf2;
  struct in_addr iptemp;

  char* vusername;
  char* vpassword;
  char* newhost;
  char* nick;

  /* Not Enough Variables */
  if (parc < 3) {
     send_reply(sptr, ERR_NEEDMOREPARAMS, "VHOST");
     return 0;
  }

  /* If FLAG_VHOST Is Already Set Deny VHOST Change */
  if (HasVHost(sptr)) {
    send_reply(sptr, ERR_HOSTALREADYSET);
    return 0;
  }

  /* Put User Defined Values Into Variables */
  nick = parv[0];
  vusername = parv[1];
  vpassword = parv[2];


  /* Check To See If vusername Is Not Over ACCOUNTLEN */
  if (strlen(vusername) > ACCOUNTLEN) {
     send_reply(sptr, ERR_VHOSTBADSIZE, "Username", "long", ACCOUNTLEN);
     return 0;
  }

  /* Check To See If vpassword Is Not Over PASSWDLEN */
  if (strlen(vpassword) > PASSWDLEN) {
     send_reply(sptr, ERR_VHOSTBADSIZE, "Password", "long", PASSWDLEN);
     return 0;
  }

  /* Find VHOST Based On Username And Password */
  if ((aconf2 = find_vhost(vusername, vpassword))!=NULL) {
    if (strcmp(vpassword, aconf2->passwd) == 0) {
      newhost = aconf2->name;
    }
  } else {
    send_reply(sptr, ERR_INVUSERORPASS);
    return 0;
  }

  /* Check If The Length Of Host Is Not Over HOSTLEN from ircd_defs.h */
  if (strlen(newhost) > HOSTLEN) {
     send_reply(sptr, ERR_VHOSTBADSIZE, "Hostname", "long", HOSTLEN);
     return 0;
  }

  /* Check To See If The Hostname Is Valid */
  if (!(IsValidHost(newhost))) {
    send_reply(sptr, ERR_BADHOST, aconf2->name);
    return 0;
  }
  /* Begin The Process Of Changing The Host If newhost Exists */
  if (newhost) {
    /* Remove Client Off All Channels */
    sendcmdto_common_channels_butone(sptr, CMD_QUIT, sptr, ":%s", 
     feature_str(FEAT_VHOST_QUITREASON));

    /* Broadcast VHOST Change To Opers */
    sendto_opmask_butone(0, SNO_OLDSNO, "%s activated user vhost for '%s@%s'", parv[0],
     cli_user(sptr)->username, newhost);

    /* Backup Original Host/IP */
    ircd_snprintf(0, cli_user(cptr)->orighost, HOSTLEN, "%s", cli_user(cptr)->host);
    ircd_snprintf(0, cli_user(sptr)->origip, SOCKIPLEN, "%s", cli_sock_ip(sptr));

    /* Change IP Over */
    iptemp.s_addr = inet_addr(feature_str(FEAT_VHOST_IP));
    cli_ip(sptr).s_addr = iptemp.s_addr;
    ircd_strncpy(cli_sock_ip(sptr), ircd_ntoa((const char*) &cli_ip(cptr)), SOCKIPLEN);

    /* Change Host Over */
    ircd_strncpy(cli_sockhost(sptr), newhost, HOSTLEN);
    ircd_strncpy(cli_user(cptr)->host, newhost, HOSTLEN);
    ircd_strncpy(cli_user(cptr)->realhost, newhost, HOSTLEN);

    /* Tell The Client His/Her New Host And Set FLAG_VHOST */
    send_reply(sptr, RPL_NEWHOSTIS, parv[0], "Your new", "is", cli_user(sptr)->username, newhost);
    SetVHost(sptr);

    /* Rejoin Client Into His/Her Channels And Set Modes */
    for (chan = cli_user(sptr)->channel; chan; chan = chan->next_channel) {
      sendcmdto_channel_butserv_butone(sptr, CMD_JOIN, chan->channel, sptr, "%H", chan->channel);
      if (IsChanOp(chan) && HasVoice(chan)) {
        sendcmdto_channel_butserv_butone(&me, CMD_MODE, chan->channel, sptr, "%H +ov %C %C",
           chan->channel, sptr, sptr);
      } else if (IsChanOp(chan) || HasVoice(chan)) {
        sendcmdto_channel_butserv_butone(&me, CMD_MODE, chan->channel, sptr, "%H +%c %C",
            chan->channel, IsChanOp(chan) ? 'o' : 'v', sptr);
      }
    }

    sendcmdto_serv_butone(sptr, CMD_VHOST, cptr, "%C %s", sptr, newhost);

  }
  return 0;
}


/*
 * IsValidHost()
 *
 * Checks if the host given to the function is valid
 * according to the DNS RFC.
 *
 */

int IsValidHost(char *host)
{
  char *s;
  char nhost[(HOSTLEN + 1)];
  
  /* Sanity check */
  if (!host)
    return -1;

  /* Copy host to a temp. array */
  ircd_strncpy(nhost,host,HOSTLEN);
 
  /* If current char. in hostmask is not alphanumeric AND isn't "-" or ".", it's
   * invalid according to DNS RFC - so disallow
   */

  for (s = nhost; *s; s++) {
    if (!isalpha(*s) && !isdigit(*s) && !(*s == '.') && !(*s == '-') && !(*s == '~')) {
       return 0;
    }
  }  

  return 1;
}

/* 
 * ms_vhost()
 * 
 * parv[0] = sender prefix
 * parv[1] = vhost target
 * parv[2] = vhost
 *
 */
int ms_vhost(struct Client *cptr, struct Client *sptr, int parc, char *parv[])
{
  struct in_addr iptemp;
  struct Membership *chan;

  char* nick;
  char* newhost;

  if (parc < 3) {
     send_reply(sptr, ERR_NEEDMOREPARAMS, "VHOST");
     return 0;
  }

  /* If FLAG_VHOST Is Already Set Deny VHOST Change */
  if (HasVHost(sptr)) {
    send_reply(sptr, ERR_HOSTALREADYSET);
    return 0;
  }

  nick = parv[1];
  newhost = parv[2];
 
  if (newhost) {
    /* Remove Client Off All Channels */
    sendcmdto_common_channels_butone(sptr, CMD_QUIT, sptr, ":%s",
     feature_str(FEAT_VHOST_QUITREASON));

    /* Broadcast VHOST Change To Opers */
    sendto_opmask_butone(0, SNO_OLDSNO, "%s activated user vhost for '%s@%s'", parv[0],
     cli_user(sptr)->username, newhost);

    /* Backup Original Host/IP */
    ircd_snprintf(0, cli_user(sptr)->orighost, HOSTLEN, "%s", cli_user(sptr)->host);
    ircd_snprintf(0, cli_user(sptr)->origip, SOCKIPLEN, "%s", cli_sock_ip(sptr));

    /* Change IP Over */              
    iptemp.s_addr = inet_addr(feature_str(FEAT_VHOST_IP));             
    cli_ip(sptr).s_addr = iptemp.s_addr;
    ircd_strncpy(cli_sock_ip(sptr), ircd_ntoa((const char*) &cli_ip(sptr)), SOCKIPLEN);


    /* Change Host Over */
    ircd_strncpy(cli_sockhost(sptr), newhost, HOSTLEN);   
    ircd_strncpy(cli_user(sptr)->host, newhost, HOSTLEN);
    ircd_strncpy(cli_user(sptr)->realhost, newhost, HOSTLEN);

    SetVHost(sptr);

    /* Rejoin Client Into His/Her Channels And Set Modes */
    for (chan = cli_user(sptr)->channel; chan; chan = chan->next_channel) {
      sendcmdto_channel_butserv_butone(sptr, CMD_JOIN, chan->channel, sptr, "%H", chan->channel);
      if (IsChanOp(chan) && HasVoice(chan)) {
        sendcmdto_channel_butserv_butone(&me, CMD_MODE, chan->channel, sptr, "%H +ov %C %C",
           chan->channel, sptr, sptr);
      } else if (IsChanOp(chan) || HasVoice(chan)) {
        sendcmdto_channel_butserv_butone(&me, CMD_MODE, chan->channel, sptr, "%H +%c %C",
            chan->channel, IsChanOp(chan) ? 'o' : 'v', sptr);
      }
    }

    sendcmdto_serv_butone(sptr, CMD_VHOST, cptr, "%C %s", sptr, newhost);  
  }
 
 return 0;
}




